//
//  altoweb.cpp
//  restore_alto_files
//
//  Created by Paul McJones starting 10/30/13.
//  Copyright (c) 2013 Paul McJones. All rights reserved.
//

#include <cassert>
#include <fstream>
#include <memory>
#include <string>
//#include <unistd.h>
#include <vector>
#include "altoweb.h"
#include "bfs.h"
#include "bravo2html.h"
#include "press2pdf.h"
#include "tioga.h"
#include "TiogaOutput.h"
#include "pstream.h"
#include "utility.h"
#include "web.h"

namespace altoweb {
    
    typedef std::string Name; // letter, digit, period, ...
    typedef Name ServerName;
    typedef Name DirectoryName;
    typedef std::string DirPath; // DirectoryName (">" DirectoryName)...
    typedef Name FileName;
    typedef int  Version;
    typedef Name Author;
    typedef uint64_t Crc32;
    
    typedef ci_less name_less;
    
    struct Directory;
    
    typedef std::map<DirectoryName, std::shared_ptr<Directory>, name_less> DirectoryDirs;
    
    typedef DirectoryDirs::iterator         DirectoryDirsIterator;
    typedef DirectoryDirs::const_iterator   DirectoryDirsConstIterator;
    
    typedef struct {
        time_t      creation_date;  // e.g., 31-Jan-81 19:59:27 PST
        time_t      write_date;     //       (same format)
        time_t      read_date;      //       (same format)
        file_type   f_t;
        std::fstream::streamoff size;   // in bytes
        Author      author;         // e.g., Murray
        Crc32       crc32;
        RecordIdentifier    ri;
    } File;

    // Weak ordering for DirectoryFiles map
    struct name_version_less {
        name_version_less() {}
        bool operator()(const std::pair<FileName, Version>& a,
                        const std::pair<FileName, Version>& b) const {
            if (name_less()(a.first, b.first)) return true;
            if (name_less()(b.first, a.first)) return false;
            return a.second < b.second;
        }
    };
    
    typedef std::map<std::pair<FileName, Version>, File, name_version_less> DirectoryFiles;
    
    typedef DirectoryFiles::iterator        DirectoryFilesIterator;
    typedef DirectoryFiles::const_iterator  DirectoryFilesConstIterator;
    
    struct Directory {
        DirectoryDirs directories;
        DirectoryFiles files;
    };
    
    std::map<ServerName, Directory, name_less> server; // but no files at top level
    
    typedef std::map<ServerName, Directory, name_less>::iterator        ServerIterator;
    typedef std::map<ServerName, Directory, name_less>::const_iterator  ServerConstIterator;
    
    // Ensure that the directories from sn through dp exist, returning pointer to the last one
    /* An IFS server has only directories at the root, but an "extra" server corresponding to
       a CD-ROM image may have files at the root. */
    Directory* make_path(ServerName sn, DirPath dp) {
        // TODO: test for presence first, and only call insert when necessary
        std::pair<ServerIterator, bool> s_pair = server.insert(std::make_pair(sn, Directory()));
        Directory* dir = &(s_pair.first->second);
        std::string::size_type i = 0;
        std::string::size_type j;
        std::string::size_type n = dp.length();
        while (i != n) {
            j = dp.find(">", i);
            std::string dir_name(dp.substr(i, j == std::string::npos ? j : j - i));
            // TODO: test for presence first, and only call insert when necessary
            /* ***** DEBUG */ if (dir_name.empty())
                std::cout << "***** Inserting empty directory name into model\n";
            std::pair<DirectoryDirsIterator, bool> d_pair
                = dir->directories.insert(std::make_pair(dir_name,
                                                         std::shared_ptr<Directory>(new Directory())));
            dir = &(*d_pair.first->second);
            if (j != std::string::npos)
                i = j + 1;
            else
                i = n;
        }
        return dir;
    }

    void add_file(const FileRecord& f) {
        /* ***** DEBUG */if (f.dir_rec.file_name.empty())
            std::cout << "***** Empty file_name being added.\n";
        Directory* dir = make_path(f.dir_rec.server, f.dir_rec.dir_path);
        std::pair<DirectoryFilesIterator, bool> d_pair =
            dir->files.insert(std::make_pair(std::make_pair(f.dir_rec.file_name, f.dir_rec.version), File()));
        File& file = d_pair.first->second;
        file.creation_date = f.dir_rec.creation_date;
        file.write_date = f.dir_rec.write_date;
        file.read_date = f.dir_rec.read_date;
        file.f_t = f.data_rec.f_t;
        file.size = f.dir_rec.size;
        file.author = f.dir_rec.author;
        file.crc32 = f.data_rec.crc32;
        file.ri = f.ri;
    }
    
    // Weak ordering on file records, for building web model.
    bool less_by_full_path(const FileRecord& a, const FileRecord& b) {
        if (ci_less()(a.dir_rec.server, b.dir_rec.server))          return true;
        if (ci_less()(b.dir_rec.server, a.dir_rec.server))          return false;
        if (ci_less()(a.dir_rec.dir_path, b.dir_rec.dir_path))      return true;
        if (ci_less()(b.dir_rec.dir_path, a.dir_rec.dir_path))      return false;
        if (ci_less()(a.dir_rec.file_name, b.dir_rec.file_name))    return true;
        if (ci_less()(b.dir_rec.file_name, a.dir_rec.file_name))    return false;
        if (a.dir_rec.version < b.dir_rec.version)                  return true;
        if (b.dir_rec.version < a.dir_rec.version)                  return false;
        return false;
    }
    
    void build_model(std::vector<FileRecord>& file_table) {
        std::sort(file_table.begin(), file_table.end(), less_by_full_path);
        auto f = file_table.cbegin();
        auto l = file_table.cend();
        while (f != l) {
            add_file(*f);
            ++f;
        }
    }

    // Generate a page with a (partial) octal dump of the file contents.
    void gen_dump(std::ofstream& os, const std::string& file_path, const std::string& base_relative_file_path) {
        // Consider endianness if displaying more than a byte at a time.
        //std::string command_prefix = "od -t oC -N 100000 "; // octal short; maximum of 100000 bytes
        std::string command_prefix = "od -c -N 50000 "; // C-style escaped characters; maximum of 50000 bytes
        os << "$ " << command_prefix << base_relative_file_path << std::endl;
        std::string command = command_prefix + shell_quote(file_path);
        
        redi::ipstream is(command);
        if (is.fail() || is.eof())
            std::cout << "*** gen_dump: od command failed to produce output for " << file_path << "\n";
        std::string line;
        while (std::getline(is, line)) {
            os << line << '\n';
        }
    }

    // Generate a page with a view of the file contents based on the file type.
    void gen_file_page(const std::string& dir_path, const std::string& base_relative_path,
                       const FileName& file_name,
                             File& file, // not const, since may set f_t = tioga
                       bool ignore_bravo,
                       bool ignore_tioga,
                       bool ignore_press)
    {
        const std::string file_path = join_paths(dir_path, file_name);
        const std::string html_path = join_paths(dir_path, "." + file_name + ".html");
        std::string alto_file_path = base_relative_path;
        auto n = alto_file_path.find('/');
        if (n == std::string::npos) {
            // CD, with top-level files -- base_relative_path is just server_name
            alto_file_path = alto_file_path + "]";
        }
        else {
            alto_file_path.replace(n, 1, "]&lt;");
            alto_file_path = alto_style_dir_name(alto_file_path);
        }
        alto_file_path = "[" + alto_file_path + file_name;
        const std::string title = "File " + alto_file_path;
        
        auto gen_ascii_contents =
            [&](std::ofstream& os, const std::string& indent) {
                std::ifstream is(file_path, std::ifstream::in | std::ifstream::binary);
                if (!is) throw raf_exception("gen_file_page could not open " + file_path);
                std::fstream::streamoff length = size(is);
                
                std::string s;
                s.resize(length);
                is.read(&s[0], length);
                s = escape_ascii(s);
                os.write(&s[0], s.size());
            };
        
        auto gen_binary_contents =
            [&](std::ofstream& os, const std::string& indent) {
                gen_dump(os, file_path, alto_file_path);
            };
        
        if (file.f_t == ascii || (file.f_t == bravo && ignore_bravo))
            gen_preformatted_page(html_path, title, gen_ascii_contents);
//        else if (file.f_t == smalltalk)
//            smalltalk_to_html(file_path, html_path);
        else if (file.f_t == bravo)
            try {
                bravo_to_html(alto_file_path, file_path, html_path);
            } catch (raf_exception p_e) {
                std::cout << "*** bravo_to_html(" << file_path << ") " << p_e.what() << std::endl;
            }
        else if (file.f_t == press && !ignore_press) {
            const std::string view_name = "." + file_name + ".pdf";
            const std::string view_path = join_paths(dir_path, view_name);
            try {
                press_to_pdf(alto_file_path, file_path, view_path, false);
            } catch (raf_exception p_e) {
                std::cout << "*** press_to_pdf(" << file_path << ") " << p_e.what() << std::endl;
            }
        }
        else if (!ignore_tioga &&
                 0 == tioga_to_text_or_html(alto_file_path.c_str(), // only for HTML title
                                            false, // generate HTML
                                            file_path.c_str(),
                                            join_paths(dir_path, "." + file_name + ".html").c_str())) {
            int err = tioga_to_text_or_html(alto_file_path.c_str(), // only for HTML title
                                            true, // generate Text
                                            file_path.c_str(),
                                            join_paths(dir_path, "." + file_name + ".txt").c_str());
            /* ***** DEBUG */ if (err != 0)
                std::cout << "tioga_to_text failed on " << file_path << "\n";
                
            file.f_t = tioga;
        }
        else // binary
            gen_preformatted_page(html_path, title, gen_binary_contents);
    }
    
    void gen_logo(std::ofstream&os) {
        gen_div(os, "", [](std::ofstream& os, const std::string& indent) {
            // *** Should have gen_image; gen_anchor should take a contents function.
            gen_anchor(os, "http://www.computerhistory.org",
                       "<img src=\"https://computerhistory.org/wp-content/themes/bsdstarter/assets/img/chm-logo-text.svg\" title=\"Computer History Museum Logo\" alt=\"Computer History Museum Logo\" width=\"108px\">");
        });
    }
    
    const bool ordered = true;
    const std::string style_monospaced = "style=\"font: 10pt monospace\"";
    
    // Generate a page with links to each subdirectory and to each file
    // and recurse for each subdirectory.
    void gen_dir_page(const std::string& dir_path,
                      const std::string& base_relative_path,
                      const DirectoryName& dir_name,
                      Directory& dir,
                      bool preserve_case, bool ignore_bravo, bool ignore_tioga,
                      bool ignore_press) {
        // ***** title should be full path with "breadcrumb" links
        const std::string base_relative_dir_path = join_paths(base_relative_path, dir_name);
        std::string server_and_dir_path = base_relative_dir_path;
        server_and_dir_path.replace(server_and_dir_path.find('/'), 1, "]&lt;");
        const std::string title = "Directory [" + alto_style_dir_name(server_and_dir_path);
        
        // For each subdirectory, recurse
        auto f_dirs = dir.directories.begin();
        auto l_dirs = dir.directories.end();
        while (f_dirs != l_dirs) {
            const DirectoryName subdir_name = f_dirs->first;
            // If the last character of the subdir_name is '_', this is
            // an expanded DM or image file, so we always preserve the case
            bool expanded_file = subdir_name.back()=='_';
            std::string subdir_path = join_paths(dir_path, (preserve_case||expanded_file)?subdir_name:force_lower_case(subdir_name));
            Directory& subdir = *f_dirs->second;
            mkpath_np(subdir_path.c_str(), 0777);
            gen_dir_page(subdir_path, base_relative_dir_path, subdir_name, subdir, preserve_case, ignore_bravo, ignore_tioga, ignore_press);
            ++f_dirs;
        }
        
        // For each file, generate a viewable proxy if possible
        auto f_files = dir.files.begin(); // not const, in case ft_t must be set
        auto l_files = dir.files.end();
        while (f_files != l_files) {
            const Version version = f_files->first.second;
            const FileName file_name = append_version(f_files->first.first, version);
            File& file = f_files->second; // not const, in case f_t must be set
            gen_file_page(dir_path, base_relative_dir_path, file_name, file, ignore_bravo, ignore_tioga, ignore_press);
            ++f_files;
        }

        auto gen_dir_contents = [&](std::ofstream& os, const std::string& indent) {
            gen_logo(os);
            
            gen_heading(os, 1, title);
            
            os << indent; gen_heading(os, 2, "Subdirectories");

            DirectoryDirsConstIterator f_dirs = dir.directories.cbegin();
            DirectoryDirsConstIterator l_dirs = dir.directories.cend();
            std::function<void (DirectoryDirsConstIterator f)> gen_dir_item_contents = [&os,preserve_case](DirectoryDirsConstIterator f) {
                    DirectoryName subdir_name = f->first;
                    // If the last character of the subdir_name is '_', this is
                    // an expanded DM or image file, so we always preserve the case
                    bool expanded_file = subdir_name.back()=='_';
                    gen_anchor(os,
                               join_paths(
                                    (preserve_case||expanded_file)?subdir_name:force_lower_case(subdir_name),
                                    ".index.html"),
                               alto_style_dir_name(subdir_name));
                };
            gen_list(os, indent, f_dirs, l_dirs, gen_dir_item_contents, ordered, style_monospaced);
            
            os << indent; gen_heading(os, 2, "Files");
            
            DirectoryFilesConstIterator f_files = dir.files.cbegin();
            DirectoryFilesConstIterator l_files = dir.files.cend();
            std::function<void (DirectoryFilesConstIterator f)> gen_file_item_contents = [&os, ignore_bravo, ignore_tioga, ignore_press](DirectoryFilesConstIterator f) {
                const Version version = f->first.second;
                const FileName file_name = append_version(f->first.first, version);
                const File& file = f->second;
                const std::string view_name = "."
                    + file_name
                    + (file.f_t == press && !ignore_press? ".pdf" : ".html");
                gen_anchor(os, view_name, file_name);
                os << " " << file_type_to_string(file.f_t) << " ";
                if (file.f_t == tioga) {
                    const std::string text_view_name = "."
                        + file_name
                        + ".txt";
                    gen_anchor(os, text_view_name, "(Text)");
                    os << " ";
                }
                gen_anchor(os, file_name, "(Original)");
                os << " ";
                os << file.size
                << " " << date_to_string(file.creation_date)
                << " " << file.author;
            };
            gen_list(os, indent, f_files, l_files, gen_file_item_contents, ordered, style_monospaced);
        };
        
        // Generate the page representing the directory
        std::cout << "Generating " << base_relative_dir_path << "\n";
        gen_page(join_paths(dir_path, ".index.html"), title, gen_dir_contents);
    }
    
    // Generate a page with links to each directory on given server.
    void gen_server_page(const std::string& web_path, ServerIterator f_server,
                         bool preserve_case, bool ignore_bravo, bool ignore_tioga,
                         bool ignore_press) {
        const ServerName& server_name = f_server->first;
        Directory& dir = f_server->second;
        const std::string title = "Server [" + server_name + "]";
        const std::string server_path =
            join_paths(web_path, preserve_case?server_name:force_lower_case(server_name));
        mkpath_np(server_path.c_str(), 0777);

        
        // For each directory, generate a page with links to each subdirectory and to each file
        auto f_dirs = dir.directories.begin();
        auto l_dirs = dir.directories.end();
        while (f_dirs != l_dirs) {
            const DirectoryName& dir_name = f_dirs->first;
            const std::string dir_path = join_paths(server_path, preserve_case?dir_name:force_lower_case(dir_name));
                mkpath_np(dir_path.c_str(), 0777);
            gen_dir_page(dir_path, server_name, dir_name, *f_dirs->second, preserve_case, ignore_bravo, ignore_tioga, ignore_press);
            ++f_dirs;
        }
        
        // For each file, generate a viewable proxy if possible
        auto f_files = dir.files.begin(); // not const, in case ft_t must be set
        auto l_files = dir.files.end();
        while (f_files != l_files) {
            const Version version = f_files->first.second;
            const FileName file_name = append_version(f_files->first.first, version);
            File& file = f_files->second; // not const, in case f_t must be set
            gen_file_page(server_path, server_name, file_name, file, ignore_bravo, ignore_tioga, ignore_press);
            ++f_files;
        }
        
        gen_page(join_paths(server_path, ".index.html"),
                 title,
                 [&, f_dirs, l_dirs](std::ofstream& os, const std::string& indent) {
                     gen_logo(os);
                     gen_heading(os, 1, title);
                     gen_heading(os, 2, "Directories");
                     std::function<void (DirectoryDirsIterator f)> gen_dir_item_contents
                        = [&os,preserve_case](DirectoryDirsIterator f) {
                            const DirectoryName dir_name = f->first;
                            // These are top-level directories, so begin each name with "<" (escaped for HTML as "&lt;").
                            std::string dir_path(
                                preserve_case?dir_name:force_lower_case(dir_name));
                            gen_anchor(os, join_paths(dir_path, ".index.html"), "&lt;" + alto_style_dir_name(dir_name));
                        };
                     gen_list(os, indent, dir.directories.begin(), l_dirs, gen_dir_item_contents, ordered, style_monospaced);
            
                    DirectoryFilesConstIterator f_files = dir.files.cbegin();
                    DirectoryFilesConstIterator l_files = dir.files.cend();
                    if (f_files != l_files) {
                        gen_heading(os, 2, "Files");
                        std::function<void (DirectoryFilesConstIterator f)> gen_file_item_contents = [&os, ignore_bravo, ignore_tioga, ignore_press](DirectoryFilesConstIterator f) {
                            const Version version = f->first.second;
                            const FileName file_name = append_version(f->first.first, version);
                            const File& file = f->second;
                            const std::string view_name = "."
                                + file_name
                                + (file.f_t == press && !ignore_press? ".pdf" : ".html");
                            gen_anchor(os, view_name, file_name);
                            os << " " << file_type_to_string(file.f_t) << " ";
                            gen_anchor(os, file_name, "(raw)");
                            os << " ";
                            os << file.size
                            << " " << date_to_string(file.creation_date)
                            << " " << file.author;
                        };
                        gen_list(os, indent, f_files, l_files, gen_file_item_contents, ordered, style_monospaced);
                   }
                 });
    }

    std::string extension_frequencies_name = "extension_frequencies.html";
    std::string author_frequencies_name = "author_frequencies.html";
    std::string cross_ref_name = "cross_reference.html";
    std::string conversion_log_name = "conversion_log.txt";
    
    template <typename T, typename V, typename R>
    //    requires(Regular(T) && Regular(V) && StrictWeakOrder(R))
    void gen_frequencies_page(const std::string& web_path, const std::map<T, V, R>& table, const std::string& title) {
        auto gen_contents = [&](std::ofstream& os, const std::string& indent){
            gen_logo(os);
            gen_heading(os, 1, title);
            typedef typename std::map<T, V, R>::const_iterator I;
            gen_list<I>(os, indent,
                        table.cbegin(), table.cend(),
                        [&](I f) -> void {os << indent << "  " << f->first << " " << f->second << "\n";}, ordered, style_monospaced);
        };
        gen_page(web_path, title, gen_contents);
    }
    
    // Ordering on file records, for print_cross_ref.
    bool less_by_file_name(const FileRecord& a, const FileRecord& b) {
        if (ci_less()(a.dir_rec.file_name, b.dir_rec.file_name))    return true;
        if (ci_less()(b.dir_rec.file_name, a.dir_rec.file_name))    return false;
        if (ci_less()(a.dir_rec.server, b.dir_rec.server))          return true;
        if (ci_less()(b.dir_rec.server, a.dir_rec.server))          return false;
        if (ci_less()(a.dir_rec.dir_path, b.dir_rec.dir_path))      return true;
        if (ci_less()(b.dir_rec.dir_path, a.dir_rec.dir_path))      return false;
        if (a.dir_rec.version < b.dir_rec.version)                  return true;
        if (b.dir_rec.version < a.dir_rec.version)                  return false;
        return false;
    }
    
    void gen_cross_reference(const std::string& web_path, std::vector<FileRecord>& file_table,
                             bool preserve_case) {
        const std::string title = "Cross reference by file name";
        auto gen_contents = [&](std::ofstream& os, const std::string& indent){
            gen_logo(os);
            gen_heading(os, 1, title);
            os << indent << "<ul " << style_monospaced << ">\n";
            os << indent << "  <li>file_name\n";
            os << indent << "    <ul><li>full_name length crc32 creation (date time) file_type author archive_id(year tape record)</li></ul>\n";
            os << indent << "  </li>\n";
            os << indent << "</ul>\n";
            auto f = file_table.cbegin();
            auto l = file_table.cend();
            os << indent << "<ul " << style_monospaced << ">\n";
            while (f != l) {
                os << indent << "  <li>";
                os << f->dir_rec.file_name << "\n";
                os << indent << "    <ul>\n";
                auto first = f;
                do {
                    std::string web_dir_path = join_paths(join_paths(f->dir_rec.server,
                                                                     subst(f->dir_rec.dir_path, '>', '/')),
                                                          ".index.html");
                    if (!preserve_case) web_dir_path = force_lower_case(web_dir_path);
                    os  << indent << "      <li>"
                        << "<a href=\"" << web_dir_path << "\">"
                            << "[" << f->dir_rec.server << "]&lt;" << f->dir_rec.dir_path
                        << "&gt;</a>"
                        << append_version(f->dir_rec.file_name, f->dir_rec.version) << " "
                        // end of sort key
                        << f->dir_rec.size << " "
                        << std::hex << f->data_rec.crc32 << std::dec << " "
                        << date_to_string(f->dir_rec.creation_date) << " "
                        << file_type_to_string(f->data_rec.f_t) << " "
                        << f->dir_rec.author << " "
                        << std::get<0>(f->ri) << " " << std::get<1>(f->ri) << " " << std::get<2>(f->ri)
                        << "</li>\n";
                    ++f;
                } while (f != l && ci_equal(f->dir_rec.file_name, first->dir_rec.file_name));
                os << indent << "    </ul>\n";
                os << indent << "  </li>\n";
            }
            os << indent << "</ul>\n";
        };
        
        std::sort(file_table.begin(), file_table.end(), less_by_file_name);
        gen_page(join_paths(web_path, cross_ref_name), title, gen_contents);
    }
    
    // Generate a top-level page with a list of links to servers.
    void gen_top_level_page(const std::string& web_path,
                            std::vector<FileRecord>& file_table,
                            std::multimap<Crc32, std::vector<FileRecord>::size_type>file_table_crc32_index,
                            std::map<Extension, int, ci_less>extension_table,
                            std::map<Author, int, ci_less> author_table, bool preserve_case) {
        
        //const std::string title = "Xerox PARC Alto filesystem archive";
        const std::string title = "Xerox PARC Interim File System (IFS) archive";
        const std::string tracker =
            "<!-- Google Tag Manager -->\n"
            "<script>(function(w,d,s,l,i){w[l]=w[l]||[];w[l].push({'gtm.start':\n"
            "new Date().getTime(),event:'gtm.js'});var f=d.getElementsByTagName(s)[0],\n"
            "j=d.createElement(s),dl=l!='dataLayer'?'&l='+l:'';j.async=true;j.src=\n"
            "'https://www.googletagmanager.com/gtm.js?id='+i+dl;f.parentNode.insertBefore(j,f);\n"
            "})(window,document,'script','dataLayer','GTM-MLN7JPS');</script>\n"
            "<!-- End Google Tag Manager -->\n";
        auto gen_top_level_contents =
        [&](std::ofstream& os, const std::string& indent) {
            gen_logo(os);
            
            gen_heading(os, 1, title);
            
            // *** Should use gen_anchor; gen_paragraph should take a contents function.
            
            gen_paragraph(os, "", "With the permission of <a href=\"http://www.parc.com\">PARC</a> (formerly the Xerox Palo Alto Research Center, now a part of SRI International), the <a href=\"http://www.computerhistory.org/\">Computer History Museum</a> is pleased to make available, for non-commercial use only, snapshots of source code, executables, documentation, font files, and other files. The files are organized by the original server on which they resided at PARC. For each server, there are one or more top-level directories containing subdirectories and files. Each file is generally presented two ways: in a viewable format (e.g., HTML, PDF), and in its original format (referred to as \"raw\" in the directory listings).</p><p>We invite you to explore the archive itself. You can do that by browsing the contents of each of the PARC IFS servers from the list below, or by doing text searches within the single large page <a href=\"cross_reference.html\">Cross-reference by name</a> (warning: it takes a long time to load).</p><p>Alternatively, you may prefer to explore the archive by viewing the <a href=\"Xerox_PARC_source_code.html\">walkthrough</a> page, which provides more context on the various projects and programs, who wrote them, links to references, and details about the archive, including provenance, naming conventions, file types, etc.</p><p>You can read more about the story of this digital archive, its preservation and recovery, and a sample of what interesting discoveries can be made in it at this blog post, <a href=\"https://computerhistory.org/blog/a-backup-of-historical-proportions/\">\"A Backup of Historical Proportions,\"</a> by CHM curator David Brock.");
            
            gen_heading(os, 2, "Servers");
            
            // For each server
            gen_list<ServerConstIterator>(os, indent, server.cbegin(), server.cend(),
                                          [&os,preserve_case](ServerConstIterator f) -> void {
                                              const ServerName& server_name = f->first;
                                              std::string server_path(preserve_case?server_name:force_lower_case(server_name));
                                              gen_anchor(os, join_paths(server_path, ".index.html"), "[" + server_name + "]");
                                          }, ordered, style_monospaced);
            
            gen_heading(os, 2, "Summaries");
            gen_paragraph(os, "", std::to_string(file_table.size()) + " total files; "
                                  + std::to_string(key_count(file_table_crc32_index)) + " unique crc32s");
            
            gen_anchor(os, cross_ref_name, "Cross-reference by name", para);
            gen_anchor(os, extension_frequencies_name, "Extension frequencies", para);
            gen_anchor(os, author_frequencies_name, "Author frequencies", para);
            gen_anchor(os, conversion_log_name, "Conversion log", para);
        };
        gen_page(join_paths(web_path, "index.html"),
                 title,
                 gen_top_level_contents,
                 tracker);
    }
    
    void gen_pages(const std::string& web_path,
                   std::vector<FileRecord>& file_table,
                   std::multimap<Crc32, std::vector<FileRecord>::size_type>file_table_crc32_index,
                   std::map<Extension, int, ci_less>extension_table,
                   std::map<Author, int, ci_less> author_table, bool preserve_case,
                   bool ignore_bravo, bool ignore_tioga, bool ignore_press) {
        std::cout << "\nGenerate web pages." << std::endl;

        mkpath_np(web_path.c_str(), 0777);
        
        build_model(file_table);
        
        // Generate a page with a link to each server.
        gen_top_level_page(web_path, file_table, file_table_crc32_index, extension_table, author_table, preserve_case);
        
        // For each server, generate a page with a link to each directory.
        auto f_server = server.begin();
        auto l_server = server.end();
        while (f_server != l_server) {
            gen_server_page(web_path, f_server, preserve_case, ignore_bravo, ignore_tioga, ignore_press);
            ++f_server;
        }
        
        // Generate summary pages.
        std::cout << "\nGenerating frequency tables.\n";
        gen_frequencies_page(join_paths(web_path, extension_frequencies_name), extension_table, "Extension frequencies");
        gen_frequencies_page(join_paths(web_path, author_frequencies_name), author_table, "Author frequencies");
        std::cout << "\nGenerating cross reference.\n"
                  << "  " << file_table.size() << " entries including DM and disk image contents." << std::endl;
        gen_cross_reference(web_path, file_table, preserve_case);
    }
}

