// Written by Josh Dersch, summer 2021

/*

TiogaOutput.cpp --

    Implementations of ITiogaOutput for plaintext and HTML conversion.
    Limitations:
        - Text output is extremely simplistic and has too much hard-coded behavior.

        - HTML output depends on hand-coded "template.html" for CSS templates implementing
          Tioga Styles.
        - Since CSS templates are fixed at runtime, some of the more dynamic Tioga Style
          functionality isn't possible (selecting style options based on conditionals, for
          example).
        - Would probably be worthwhile to integrate a JaM parser to allow for more accurate
          (and automatic) conversion from Tioga style definitions.
        - Greek, Japanese, Math, and special "XEROX Font" "looks" are not handled at all.
          These will require special handling for translating XCC codes to Unicode.

        - Properties (HandleProps) are not handled at all.

        - There is no special handling for the various table styles.

        - Comments are always enclosed in square brackets for text, and italicized in HTML.
*/


#include <stdio.h>
#include <stdlib.h>
#include <cstring>
//#include <string> // included in TiogaOutput.h
#include <cassert>
//#include <vector> // included in TiogaOutput.h
#include "tioga.h"
#include "tread.h"
#include "TiogaOutput.h"
#include "get_template.h"

/*

TiogaOutputText - Dead-simple raw text output.

*/

TiogaOutputText::TiogaOutputText(FILE* output, unsigned columnWidth) :
    ITiogaOutput(output),
    m_columnWidth(columnWidth),
    m_currentIndent(0)
{

}

TiogaOutputText::~TiogaOutputText()
{

}

void TiogaOutputText::StartNode(tread_Reader* reader, const char* format, const char* comment)
{
    ITiogaOutput::StartNode(reader, format, comment);

//    WriteOutput(L"\n");

    // Handle a few node formats that are applicable to plaintext (namely: indentation):
//    if (format && (_strcmpi(format, "body") == 0 ||
//                   _strcmpi(format, "body1") == 0 ||
//                   _strcmpi(format, "indent") == 0 ||
//                   _strcmpi(format, "indent1") == 0))
    if (format && (strcasecmp(format, "body") == 0 ||
                   strcasecmp(format, "body1") == 0 ||
                   strcasecmp(format, "indent") == 0 ||
                   strcasecmp(format, "indent1") == 0 ||
                   strcasecmp(format, "code") == 0))
    {
        unsigned indent = 4;
        m_indentStack.push(indent);
        m_currentIndent += indent;
    }
    else
    {
        // No indent
        m_indentStack.push(0);
    }
}

void TiogaOutputText::EndNode(tread_Reader* reader)
{

    unsigned indent = m_indentStack.top();
    m_indentStack.pop();
    m_currentIndent -= indent;

    assert(m_currentIndent >= 0);
}

void TiogaOutputText::InsertText(tread_Reader* reader, const unsigned char* text, long len, int comment)
{
//    assert(len >= m_currentLooks.size());
    if (len < m_currentLooks.size())
        tioga_Error("more looks than text");
    uint32_t lastLook = 0;

    unsigned column = 0;

    std::wstring currentLine;
    std::wstring currentWord;

    if (comment)
    {
        currentWord.append(L"<<");
        column += 2;
    }

    for (int i = 0; i < len; i++)
    {
        // If we've hit the end of the column,
        // output what we have and continue with the current word
        // we are in the middle of.
        if (column >= m_columnWidth)
        {
            column = m_currentIndent + currentWord.length();
            WriteIndent();
//            WriteOutput(L"%s\n", currentLine.c_str());
            WriteOutput(L"%ls\n", currentLine.c_str());
            currentLine.clear();
        }

        //
        // Check look: if look has changed since last character then we need to mark the change in the
        // output.
        // TODO: this logic should be factored out.
        //
        uint32_t newLook = i < m_currentLooks.size() ? m_currentLooks[i] : 0;

        if (lastLook != newLook)
        {
            ChangeLook(lastLook, newLook);
        }

        lastLook = newLook;

        const unsigned char c = text[i];
        
        // Normal look handling
        switch (c)
        {
        case '\n':
        case '\r':
        case '\t':
        case ' ':
        {
            // If we have any characters in the word, whitespace marks the end of the word;
            // insert the word and start a new word
            if (currentWord.length() > 0)
            {
                currentLine.append(currentWord);
                currentWord.clear();
            }

            // TODO: what if a tab goes over the column limit (do we care?)
            if (c == '\t')
            {
                currentLine.append(L"    ");
                column += 4;
            }
            else if (c == '\r' || c == '\n')
            {
                WriteIndent();
//                WriteOutput(L"%s\n", currentLine.c_str());
                WriteOutput(L"%ls\n", currentLine.c_str());
//                for (size_t i = 0; i <= currentLine.length(); ++i)
//                    WriteOutput(L"%lc", currentLine[i]);
                currentLine.clear();
                column = m_currentIndent;
            }
            else
            {
                currentLine.push_back(L' ');
                column++;
            }
        }
        break;

        default:
            if (newLook & (1 << (31 - ('m' - 'a'))))
            {
                // Math char
                currentWord.push_back(ConvertMathChar(c));
            }
            else if (newLook & (1 << (31 - ('g' - 'a'))))
            {
                // Greek char
                currentWord.push_back(ConvertGreekChar(c));
            }
            else
            {
                currentWord.push_back(static_cast<wchar_t>(c));
            }
            column++;
        }
    }

    // Flush the last word
    currentLine.append(currentWord);

    if (currentLine.length() > 0)
    {
        WriteIndent();
//        WriteOutput(L"%s", currentLine.c_str());
        WriteOutput(L"%ls", currentLine.c_str());
    }

    if (comment) 
    {
        WriteOutput(L">>");
    }

    // Clear all currently-applied looks.
    ChangeLook(lastLook, 0);

    WriteOutput(L"\n");
}

void TiogaOutputText::HandleProp(tread_Reader* reader, const char* propName, const unsigned char* prop, long len)
{

}

void TiogaOutputText::WriteIndent()
{
    for (unsigned i = 0; i < m_currentIndent; i++)
    {
        WriteOutput(L" ");
    }
}


/*

TiogaOutputHtml - Slightly-less simple HTML output.

*/

TiogaOutputHtml::TiogaOutputHtml(FILE* output) :
    ITiogaOutput(output)
{

}

TiogaOutputHtml::~TiogaOutputHtml()
{

}

void TiogaOutputHtml::BeginOutput(const char* title)
{
    // Inject the CSS template containing style rules into the output:
//    FILE* file = nullptr;
//    errno_t err = fopen_s(&file,"template.html", "r");
//    file = fopen("/Users/pmcjones/Documents/Xcode/restore_alto_files/tioga/template.html", "r");

//    char buffer[2048];

//    while (size_t read = fread(reinterpret_cast<void*>(buffer), sizeof(char), sizeof(buffer) - 1, file))
//    {
//        buffer[read] = '\0';
// //       WriteOutput(L"%S", buffer);
//        WriteOutput(L"%s", buffer);
//    }
    WriteOutput(L"%s", get_template());
    
//    WriteOutput(L"\n</body>\n");
    WriteOutput(L"<title>%s</title>\n</head>\n<body>\n", title);

//    fclose(file);
}

void TiogaOutputHtml::StartNode(tread_Reader* reader, const char* format, const char* comment)
{
    ITiogaOutput::StartNode(reader, format, comment);
    if (format == nullptr)
    {
        // Unsure at this time why some nodes come in with a null format:
        // is this a bug in the viewtioga code, or is this expected. 
        // If the latter, what should we default to?
        format = "code";
    }
//    WriteOutput(L"<div class=\"%S\">\n", format);
    WriteOutput(L"<div class=\"%s\">\n", format);
}

void TiogaOutputHtml::EndNode(tread_Reader* reader)
{
    const char* tag = nullptr;

    WriteOutput(L"</div>\n");

}

void TiogaOutputHtml::InsertText(tread_Reader* reader, const unsigned char* text, long len, int comment)
{
//    assert(len >= m_currentLooks.size());
    if (len < m_currentLooks.size())
        tioga_Error("more looks than text");

    uint32_t lastLook = 0;
    if (comment)
    {
        // Ensure comments are italicized.
        // This should really be done by applying styles but this cannot yet be done conditionally.
        WriteOutput(L"<i>");
    }

    for (int i = 0; i < len; i++)
    {
        //
        // Check look: if look has changed since last character then we need to mark the change in the
        // output.
        // TODO: this logic should be factored out.
        //
        uint32_t newLook = i < m_currentLooks.size() ? m_currentLooks[i] : 0;

        if (lastLook != newLook)
        {
            ChangeLook(lastLook, newLook);
        }

        lastLook = newLook;

        const unsigned char c = text[i];

        // Handle special font looks (TODO: is it legal for more than one of these to be set at once?)
        if (newLook & (1 << (31 - ('m' - 'a'))))
        {
            // Math font
            switch (c)
            {
            case '\n':
            case '\r':
                WriteOutput(L"<br>");
                break;

            case '\t':
                // TODO: should follow tab rules for current style...
                WriteOutput(L"&emsp;");
                break;

            default:
                WriteOutput(L"%c", ConvertMathChar(c));
                break;
            }
        }
        else
        {
            // Handle special characters.
            switch (c)
            {
            case '\n':
            case '\r':
                WriteOutput(L"<br>");
                break;

            case '\t':
                // TODO: should follow tab rules for current style...
                WriteOutput(L"&emsp;");
                break;

            case '\"':
                WriteOutput(L"&quot;");
                break;

            case '\'':
                WriteOutput(L"&apos;");
                break;

            case '&':
                WriteOutput(L"&amp;");
                break;

            case '<':
                WriteOutput(L"&lt;");
                break;

            case '>':
                WriteOutput(L"&gt;");
                break;

            case '_':
                // Left arrow
                WriteOutput(L"&#x2190");
                break;

                //case ' ':
                //    WriteOutput("&nbsp");
                //    break;

            case 0x18:
                WriteOutput(L"&mdash;");
                break;

            default:
                WriteOutput(L"%C", c);
                break;
            }
        }

    }

    // Clear all currently-applied looks.
    ChangeLook(lastLook, 0);

    if (comment)
    {
        WriteOutput(L"</i>");
    }

    WriteOutput(L"\n");
}

void TiogaOutputHtml::HandleProp(tread_Reader* reader, const char* propName, const unsigned char* prop, long len)
{
    // TODO: Anything related to Props.
}

void TiogaOutputHtml::BeginLook(uint32_t look)
{
//    assert(look < 26);
    WriteOutput(L"<span class=\"look-%C\"/>", static_cast<char>('a' + look));
}

void TiogaOutputHtml::EndLook(uint32_t look)
{
    WriteOutput(L"</span>");
}

static int StartNode(tread_Reader* reader, const char* format, const char* comment)
{
    ITiogaOutput* output = reinterpret_cast<ITiogaOutput*>(reader->data);
    output->StartNode(reader, format, comment);
    return 0;
}

static int EndNode(tread_Reader* reader)
{
    ITiogaOutput* output = reinterpret_cast<ITiogaOutput*>(reader->data);
    output->EndNode(reader);
    return 0;
}

static int AddLook(tread_Reader* reader, long looks, long start, int len)
{
    ITiogaOutput* output = reinterpret_cast<ITiogaOutput*>(reader->data);
    output->AddLook(reader, looks, start, len);
    return 0;
}

static int InsertText(tread_Reader* reader, const unsigned char* text, long len, int comment)
{
    ITiogaOutput* output = reinterpret_cast<ITiogaOutput*>(reader->data);
    output->InsertText(reader, text, len, comment);
    return 0;
}

static int HandleProp(tread_Reader* reader, const char* propName, const unsigned char* prop, long len)
{
    ITiogaOutput* output = reinterpret_cast<ITiogaOutput*>(reader->data);
    output->HandleProp(reader, propName, prop, len);
    return 0;
}

static struct tread_Procs procs = {
    StartNode,
    EndNode,
    AddLook,
    InsertText,
    HandleProp
};

int tioga_to_text_or_html(const char* ifs_file_path, // only for HTML title
                          bool textOutput,
                          const char* in_path,
                          const char* out_path)
{

    tread_Reader r = { 0 };

    FILE* inFile = nullptr;
//    errno_t err = fopen_s(&inFile, in_path, "rb");
    inFile = fopen(in_path, "rb");

    if (!inFile)
    {
        tioga_Error("*** tioga_to_text_or_html: Unable to open input file: %s", in_path);
        return errno;
    }

    FILE* outFile = nullptr;
//    err = fopen_s(&outFile, out_path, "w, ccs=UNICODE");
    outFile = fopen(out_path, "w, ccs=UNICODE");

    if (!outFile)
    {
        tioga_Error("*** tioga_to_text_or_html: Unable to create output file: %s", out_path);
 //       _fcloseall();
        int err = errno;
        fclose(inFile);
        return err;
    }

    ITiogaOutput* output = nullptr;
    
    if (textOutput)
    {
        output = new TiogaOutputText(outFile, 132);
    }
    else
    {
        output = new TiogaOutputHtml(outFile);
    }

    output->BeginOutput(ifs_file_path);
    
    int err = tioga_FromFile(&r, inFile, &procs, reinterpret_cast<void *>(output));
    
    if (err < 0)
    {
 //       tioga_Error("Error reading file %s", in_path);
        fclose(inFile);
        fclose(outFile);
        return err;
    }

    output->EndOutput();

    delete output;

//    _fcloseall();
    fclose(inFile);
    fclose(outFile);
    
    return 0;
}
